# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

import litellm
import tomlkit
import tomlkit.container
import tomlkit.exceptions
from aiconsole.api.websockets.outgoing_messages import SettingsWSMessage
from aiconsole.core.assets.asset import AssetStatus, AssetType
from aiconsole.core.project.paths import get_project_directory
from aiconsole.core.project.project import is_project_initialized
from aiconsole.utils.BatchingWatchDogHandler import BatchingWatchDogHandler
from aiconsole.utils.recursive_merge import recursive_merge
from appdirs import user_config_dir
from pydantic import BaseModel
from tomlkit import TOMLDocument
from watchdog.observers import Observer

_log = logging.getLogger(__name__)


class PartialSettingsData(BaseModel):
    code_autorun: Optional[bool] = None
    openai_api_key: Optional[str] = None
    disabled_materials: Optional[List[str]] = None
    disabled_agents: Optional[List[str]] = None
    enabled_materials: Optional[List[str]] = None
    enabled_agents: Optional[List[str]] = None
    forced_materials: Optional[List[str]] = None
    forced_agent: Optional[str] = None
    to_global: bool = False


class PartialSettingsAndToGlobal(PartialSettingsData):
    to_global: bool = False


class SettingsData(BaseModel):
    code_autorun: bool = False
    openai_api_key: Optional[str] = None
    disabled_materials: List[str] = []
    disabled_agents: List[str] = []
    enabled_materials: List[str] = []
    enabled_agents: List[str] = []
    forced_materials: List[str] = []
    forced_agent: str = ""


def _load_from_path(file_path: Path) -> Dict[str, Any]:
    with file_path.open() as file:
        document = tomlkit.loads(file.read())

    return dict(document)


def _set_openai_api_key_environment(settings: SettingsData) -> None:
    openai_api_key = settings.openai_api_key

    # This is so we make sure that it's overiden and does not env variables etc
    litellm.openai_key = openai_api_key or "invalid key"


class Settings:

    def __init__(self, project_path: Optional[Path] = None):
        self._settings = SettingsData()

        self._global_settings_file_path = Path(
            user_config_dir('aiconsole')) / "settings.toml"

        if project_path:
            self._project_settings_file_path = project_path / "settings.toml"
        else:
            self._project_settings_file_path = None

        self._observer = Observer()

        self._global_settings_file_path.parent.mkdir(
            parents=True, exist_ok=True)
        self._observer.schedule(
            BatchingWatchDogHandler(
                self.reload, self._global_settings_file_path.name),
            str(self._global_settings_file_path.parent),
            recursive=True)

        if self._project_settings_file_path:
            self._project_settings_file_path.parent.mkdir(
                parents=True, exist_ok=True)
            self._observer.schedule(
                BatchingWatchDogHandler(
                    self.reload, self._project_settings_file_path.name),
                str(self._project_settings_file_path.parent),
                recursive=True)

        self._observer.start()

    def model_dump(self) -> Dict[str, Any]:
        return self._settings.model_dump()

    def stop(self):
        self._observer.stop()

    async def reload(self, initial: bool = False):
        self._settings = await self.__load()
        await SettingsWSMessage(initial=initial).send_to_all()

    def get_asset_status(self, asset_type: AssetType, id: str) -> AssetStatus:
        s = self._settings

        if asset_type == AssetType.MATERIAL:
            if id in s.forced_materials:
                return AssetStatus.FORCED
            if id in s.disabled_materials:
                return AssetStatus.DISABLED
            return AssetStatus.ENABLED
        elif asset_type == AssetType.AGENT:
            if id == s.forced_agent:
                return AssetStatus.FORCED
            if id in s.disabled_agents:
                return AssetStatus.DISABLED
            return AssetStatus.ENABLED

        else:
            raise ValueError(f"Unknown asset type {asset_type}")

    def set_asset_status(self,
                         asset_type: AssetType,
                         id: str,
                         status: AssetStatus,
                         to_global: bool = False) -> None:

        if asset_type == AssetType.MATERIAL:
            partial_settings = {
                AssetStatus.DISABLED: PartialSettingsData(disabled_materials=[id]),
                AssetStatus.ENABLED: PartialSettingsData(enabled_materials=[id]),
                AssetStatus.FORCED: PartialSettingsData(
                    forced_materials=[id])
            }
        elif asset_type == AssetType.AGENT:
            partial_settings = {
                AssetStatus.DISABLED: PartialSettingsData(disabled_agents=[id]),
                AssetStatus.ENABLED: PartialSettingsData(enabled_agents=[id]),
                AssetStatus.FORCED: PartialSettingsData(forced_agent=id, enabled_agents=[self._settings.forced_agent] if self._settings.forced_agent else None)
            }

        else:
            raise ValueError(f"Unknown asset type {asset_type}")

        self.save(partial_settings[status],
                  to_global=to_global)

    def get_code_autorun(self) -> bool:
        return self._settings.code_autorun

    def set_code_autorun(self,
                         code_autorun: bool,
                         to_global: bool = False) -> None:
        self._settings.code_autorun = code_autorun
        self.save(
            PartialSettingsData(code_autorun=self._settings.code_autorun),
            to_global=to_global)

    async def __load(self) -> SettingsData:
        settings = {}
        paths = [self._global_settings_file_path,
                 self._project_settings_file_path]

        for file_path in paths:
            if file_path and file_path.exists():
                settings = recursive_merge(
                    settings, _load_from_path(file_path))

        forced_agents = [
            agent for agent, status in settings.get('agents', {}).items()
            if status == AssetStatus.FORCED
        ]

        settings_data = SettingsData(
            code_autorun=settings.get('settings',
                                      {}).get('code_autorun', False),
            openai_api_key=settings.get(
                'settings', {}).get('openai_api_key', None),
            disabled_materials=[
                material
                for material, status in settings.get('materials', {}).items()
                if status == AssetStatus.DISABLED
            ],
            disabled_agents=[
                agent for agent, status in settings.get('agents', {}).items()
                if status == AssetStatus.DISABLED
            ],
            forced_materials=[
                material
                for material, status in settings.get('materials', {}).items()
                if status == AssetStatus.FORCED
            ],
            forced_agent=forced_agents[0] if forced_agents else "",)

        _set_openai_api_key_environment(settings_data)

        _log.info(f"Loaded settings")
        return settings_data

    @staticmethod
    def __get_tolmdocument_to_save(file_path: Path) -> TOMLDocument:
        if not file_path.exists():
            document = tomlkit.document()
            document['settings'] = tomlkit.table()
            document['materials'] = tomlkit.table()
            document['agents'] = tomlkit.table()
            return document

        with file_path.open() as file:
            document = tomlkit.loads(file.read())
            for section in ['settings', 'materials', 'agents']:
                if section not in dict(document):
                    document[section] = tomlkit.table()

        return document

    def save(self,
             settings_data: PartialSettingsData,
             to_global: bool = False) -> None:

        if to_global:
            file_path = self._global_settings_file_path
        else:
            if not self._project_settings_file_path:
                raise ValueError(
                    "Cannnot save to project settings file, because project is not initialized")

            file_path = self._project_settings_file_path

        document = self.__get_tolmdocument_to_save(file_path)
        doc_settings: tomlkit.table.Table = document['settings']
        doc_materials: tomlkit.table.Table = document['materials']
        doc_agents: tomlkit.table.Table = document['agents']

        if settings_data.code_autorun is not None:
            doc_settings['code_autorun'] = settings_data.code_autorun

        if settings_data.openai_api_key is not None:
            doc_settings['openai_api_key'] = settings_data.openai_api_key

        if settings_data.disabled_materials is not None:
            for material in settings_data.disabled_materials:
                doc_materials[material] = AssetStatus.DISABLED.value

        if settings_data.forced_materials is not None:
            for material in settings_data.forced_materials:
                doc_materials[material] = AssetStatus.FORCED.value

        if settings_data.enabled_materials is not None:
            for material in settings_data.enabled_materials:
                doc_materials[material] = AssetStatus.ENABLED.value

        if settings_data.disabled_agents is not None:
            for agent in settings_data.disabled_agents:
                doc_agents[agent] = AssetStatus.DISABLED.value

        if settings_data.forced_agent is not None:
            doc_agents[settings_data.forced_agent] = AssetStatus.FORCED.value

        if settings_data.enabled_agents is not None:
            for agent in settings_data.enabled_agents:
                doc_agents[agent] = AssetStatus.ENABLED.value

        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open('w') as file:
            file.write(document.as_string())


async def init():
    global _settings
    _settings = Settings(get_project_directory()
                         if is_project_initialized() else None)
    await _settings.reload()


def get_aiconsole_settings() -> Settings:
    return _settings


async def reload_settings(initial: bool = False):
    global _settings
    _settings.stop()
    _settings = Settings(get_project_directory()
                         if is_project_initialized() else None)
    await _settings.reload(initial)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
from aiconsole.core.project import project


import random


def create_materials_str() -> str:
    new_line = "\n"

    enabled_materials = project.get_project_materials().enabled_assets()

    random_materials = (
        new_line.join(
            [
                f"* {c.id} - {c.usage}"
                for c in random.sample(enabled_materials, len(enabled_materials))
            ]
        )
        if enabled_materials
        else ""
    )

    return random_materials


# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
from aiconsole.consts import DIRECTOR_MIN_TOKENS, DIRECTOR_PREFERRED_TOKENS
from aiconsole.core.chat.types import Chat
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.core.gpt.request import GPTRequest
from aiconsole.core.gpt.types import GPTMessage
from aiconsole.utils.convert_messages import convert_messages
from aiconsole.api.websockets.outgoing_messages import AnalysisUpdatedWSMessage


async def gpt_analysis_text_step(
    chat: Chat,
    analysis_request_id: str,
    gpt_mode: GPTMode,
    initial_system_prompt: str,
    last_system_prompt: str,
):
    gpt_executor = GPTExecutor()

    request = GPTRequest(
        system_message=initial_system_prompt,
        gpt_mode=gpt_mode,
        messages=[*convert_messages(chat), GPTMessage(
            role="system",
            content=last_system_prompt
        )],
        presence_penalty=2,
        min_tokens=DIRECTOR_MIN_TOKENS,
        preferred_tokens=DIRECTOR_PREFERRED_TOKENS,
    )

    async for chunk in gpt_executor.execute(request):
        await AnalysisUpdatedWSMessage(
            analysis_request_id=analysis_request_id,
            agent_id=None,
            relevant_material_ids=None,
            next_step=None,
            thinking_process=gpt_executor.partial_response.choices[0].message.content,
        ).send_to_chat(chat.id)

    return gpt_executor.response.choices[0].message.content or ""



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
import json
from aiconsole.core.chat.types import AICCodeMessage, AICMessage, AICMessageGroup, Chat
from aiconsole.consts import FUNCTION_CALL_OUTPUT_LIMIT
from aiconsole.core.gpt.types import GPTFunctionCall, GPTMessage, GPTRole


from typing import List


last_system_message = None


def convert_message(group: AICMessageGroup, message: AICMessage, is_last: bool) -> List[GPTMessage]:
    global last_system_message

    result = []

    #
    # Augment the messages with system messages with meta data about which agent is speaking and what materials were available
    #

    if group.task:
        system_message = f"""
As a director I have assigned you ({group.agent_id}) and given you access to the following materials text: {", ".join(group.materials_ids) if group.materials_ids else "None"}.
""".strip()

        # Only provide a task for last message
        if is_last:
            system_message += "\n\nYour job: " + group.task

        if last_system_message != system_message:
            result.append(
                GPTMessage(
                    role="system",
                    name="director",
                    content=system_message,
                )
            )
            last_system_message = system_message

    

    if isinstance(message, AICCodeMessage):
        result.append(
            GPTMessage(
                role="assistant",
                content=message.content,
                function_call=GPTFunctionCall(
                    name="execute",
                    arguments=json.dumps(
                        {
                            "code": message.content,
                            "language": message.language,
                        }
                    ),
                ),
                name="code"
            )
        )

        for output in message.outputs:
            content = output.content

            if content == "":
                content = "No output"

            # Enforce limit on output length, and put info that it was truncated only if limit was reached, truncate so the last part remains (not the first)
            if len(content) > FUNCTION_CALL_OUTPUT_LIMIT:
                content = f"""
Output truncated to last {FUNCTION_CALL_OUTPUT_LIMIT} characters:

...
{content[-FUNCTION_CALL_OUTPUT_LIMIT:]}
""".strip()
            
            result.append(
                GPTMessage(
                    role="function",
                    content=content,
                    name="Run"
                )
            )
    else:
        result.append(
            GPTMessage(
                role=group.role,
                content=message.content,
                name=group.agent_id if group.agent_id != "user" else None
            )
        )

    return result


def convert_messages(chat: Chat) -> List[GPTMessage]:
    global last_system_message
    last_system_message = None

    messages: List[GPTMessage] = []

    for message_group in chat.message_groups:
        is_last_group = message_group == chat.message_groups[-1]
        for message in message_group.messages:
            is_last = is_last_group and message == message_group.messages[-1]
            messages.extend(
                convert_message(message_group, message, is_last=is_last)
            )

    return messages



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
from typing import List, Optional
from pydantic import BaseModel
from aiconsole.core.gpt.types import GPTChoice, GPTFunctionCall, GPTMessage, GPTResponse, GPTRole
import json
from typing import Union

def _parse_partial_json(s: str) -> Union[dict, str]:
    try:
        return json.loads(s)
    except json.JSONDecodeError:
        pass

    closing_chars_stack = []
    escaped = False
    completed_string = []

    for char in s:
        if closing_chars_stack and closing_chars_stack[-1] == '"':
            if char == '"' and not escaped:
                closing_chars_stack.pop()
            elif char == '\\':
                escaped = not escaped
            elif char == '\n' and not escaped:
                char = '\\n'
            else:
                escaped = False
        else:
            if char == '"':
                closing_chars_stack.append('"')
            elif char == '{':
                closing_chars_stack.append('}')
            elif char == '[':
                closing_chars_stack.append(']')
            elif char in ['}', ']'] and closing_chars_stack:
                closing_chars_stack.pop()

        completed_string.append(char)

    completed_string.extend(reversed(closing_chars_stack))

    try:
        return json.loads(''.join(completed_string))
    except json.JSONDecodeError:
        return s


class GPTPartialFunctionCall(BaseModel):
    name: str = ""
    arguments_builder: List[str] = []

    @property
    def arguments(self) -> Union[dict, str]:
        self.arguments_builder = [''.join(self.arguments_builder)]
        return _parse_partial_json(self.arguments_builder[0])


class GPTPartialMessage(BaseModel):
    role: Optional[GPTRole] = None
    content_builder: Union[List[str], None] = None
    
    function_call: Optional[GPTPartialFunctionCall] = None
    name: Optional[str] = None

    @property
    def content(self):
        if not self.content_builder:
            return None
    
        self.content_builder = [''.join(self.content_builder)]
        return self.content_builder[0]


class GPTPartialChoice(BaseModel):
    index: int = 0
    message: GPTPartialMessage = GPTPartialMessage()
    role: str = ""
    finnish_reason: str = ""


class GPTPartialResponse(BaseModel):
    id: str = ""
    object: str = ""
    created: int = 0
    model: str = ""
    choices: List[GPTPartialChoice] = []

    def to_final_response(self):
        return GPTResponse(
            id=self.id,
            object=self.object,
            created=self.created,
            model=self.model,
            choices=[GPTChoice(
                index=choice.index,
                message=GPTMessage(
                    role=choice.message.role or "user", # ???
                    content=choice.message.content,
                    function_call=GPTFunctionCall(
                        name=choice.message.function_call.name,
                        arguments=choice.message.function_call.arguments
                    ) if choice.message.function_call else None,
                    name=choice.message.name
                ),
                finnish_reason=choice.finnish_reason
            ) for choice in self.choices]
        )

    def apply_chunk(self, chunk: dict):

        if "id" in chunk:
            self.id = chunk["id"]

        if "error" in chunk:
            raise ValueError(chunk["error"])

        if "object" in chunk:
            self.object = chunk["object"]

        if "created" in chunk:
            self.created = chunk["created"]

        if "model" in chunk:
            self.model = chunk["model"]

        if "choices" in chunk:
            chunk_choices = chunk["choices"]

            for chunk_choice in chunk_choices:
                index = None

                if "index" in chunk_choice:
                    index = int(chunk_choice["index"])

                if index is None:
                    raise ValueError("index not found")

                if index >= len(self.choices):
                    self.choices.append(GPTPartialChoice())

                choice = self.choices[index]
                choice.index = index

                if "finish_reason" in chunk_choice:
                    choice.finnish_reason = chunk_choice["finish_reason"]

                if "role" in chunk_choice:
                    choice.role = chunk_choice["role"]

                message = choice.message

                if "delta" in chunk_choice:
                    chunk_delta = chunk_choice["delta"]

                    if "name" in chunk_delta:
                        message.name = chunk_delta["name"]

                    if "role" in chunk_delta:
                        message.role = chunk_delta["role"]

                    if "content" in chunk_delta and chunk_delta["content"] != None:
                        if message.content_builder is None:
                            message.content_builder = []

                        message.content_builder.append(chunk_delta["content"])

                    if "function_call" in chunk_delta:
                        chunk_function_call = chunk_delta["function_call"]

                        if not message.function_call:
                            message.function_call = GPTPartialFunctionCall()

                        if "name" in chunk_function_call:
                            message.function_call.name = chunk_function_call["name"]

                        if "arguments" in chunk_function_call:
                            message.function_call.arguments_builder.append(chunk_delta["function_call"]["arguments"])



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
import logging
from typing import AsyncGenerator
from aiconsole.core.assets.agents.agent import ExecutionModeContext
from aiconsole.core.code_running.code_interpreters.language_map import language_map
from aiconsole.core.execution_modes.get_agent_system_message import get_agent_system_message
from aiconsole.core.gpt.create_full_prompt_with_materials import create_full_prompt_with_materials
from aiconsole.utils.convert_messages import convert_messages
from openai_function_call import OpenAISchema
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.core.gpt.request import GPTRequest
from aiconsole.core.gpt.types import CLEAR_STR
from pydantic import Field


_log = logging.getLogger(__name__)

class python(OpenAISchema):
    """
    When you send a message containing Python code to python, it will be executed in a stateful Jupyter notebook environment
    """

    code: str = Field(
        ...,
        description="python code to execute, it will be executed in a stateful Jupyter notebook environment",
        json_schema_extra={"type": "string"}
    )

class shell(OpenAISchema):
    """
    This function executes the given code on the user's system using the local environment and returns the output.
    """

    code: str = Field(..., json_schema_extra={"type": "string"})


class applescript(OpenAISchema):
    """
    This function executes the given code on the user's system using the local environment and returns the output.
    """

    code: str = Field(..., json_schema_extra={"type": "string"})

async def execution_mode_interpreter(
    context: ExecutionModeContext,
) -> AsyncGenerator[str, None]:
    global llm

    system_message = create_full_prompt_with_materials(
        intro=get_agent_system_message(context.agent),
        materials=context.relevant_materials,
    )

    _log.debug(f"System message:\n{system_message}")

    executor = GPTExecutor()

    language = None
    code = ""

    async for chunk in executor.execute(
        GPTRequest(
            system_message=system_message,
            gpt_mode=context.agent.gpt_mode,
            messages=[message for message in convert_messages(
                context.chat)],
            functions=[
                python.openai_schema,
                shell.openai_schema,
                applescript.openai_schema
            ],
            min_tokens=250,
            preferred_tokens=2000
        )
    ):

        if chunk == CLEAR_STR:
            yield CLEAR_STR
            continue

        if ('choices' not in chunk or len(chunk['choices']) == 0):
            continue

        delta = chunk["choices"][0]["delta"]

        if "content" in delta and delta["content"]:
            yield delta["content"]

        function_call = executor.partial_response.choices[0].message.function_call

        if function_call and function_call.arguments:
            # This can now be both a string and a json object

            arguments = function_call.arguments
            languages = language_map.keys()

            if language is None and function_call.name in languages:
                # Languge is in the name of the function call
                language = function_call.name
                yield f'<<<< START CODE ({language}) >>>>'

            if isinstance(arguments, str):
                # We need to handle incorrect OpenAI responses, sometmes arguments is a string containing the code
                if arguments and not arguments.startswith("{"):
                    if language is None:
                        language = "python"
                        yield f'<<<< START CODE ({language}) >>>>'

                    code_delta = arguments[len(code):]
                    code = arguments

                    if code_delta:
                        yield code_delta
            else:
                if arguments and "code" in arguments:
                    if language is None:
                        language = "python"
                        yield f'<<<< START CODE ({language}) >>>>'
                    code_delta = arguments["code"][len(code):]
                    code = arguments["code"]

                    if code_delta:
                        yield code_delta

    if language:
        yield "<<<< END CODE >>>>"
        language = None



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from pathlib import Path
from typing import Dict, List, Optional
from aiconsole.core.assets.asset import Asset, AssetLocation, AssetStatus, AssetType
from aiconsole.core.assets.fs.delete_asset_from_fs import delete_asset_from_fs
from aiconsole.core.assets.fs.load_asset_from_fs import load_asset_from_fs
from aiconsole.core.assets.fs.move_asset_in_fs import move_asset_in_fs
from aiconsole.core.assets.fs.save_asset_to_fs import save_asset_to_fs

import watchdog.events
import watchdog.observers
from aiconsole.core.assets.materials.material import (Material)
from aiconsole.core.project.paths import get_core_assets_directory, get_project_assets_directory
from aiconsole.core.settings.project_settings import get_aiconsole_settings
from aiconsole.utils.BatchingWatchDogHandler import BatchingWatchDogHandler
from aiconsole.utils.list_files_in_file_system import list_files_in_file_system
from aiconsole.api.websockets.outgoing_messages import (ErrorWSMessage,
                                                    AssetsUpdatedWSMessage)

_log = logging.getLogger(__name__)


class Assets:
    _assets: Dict[str, Asset]

    def __init__(self, asset_type: AssetType):
        self.asset_type = asset_type
        self._assets = {}

        self.observer = watchdog.observers.Observer()

        get_project_assets_directory(asset_type).mkdir(parents=True, exist_ok=True)
        self.observer.schedule(BatchingWatchDogHandler(self.reload),
                               get_project_assets_directory(asset_type),
                               recursive=True)
        self.observer.start()

    def stop(self):
        self.observer.stop()

    def all_assets(self) -> List[Asset]:
        """
        Return all loaded assets.
        """
        return list(self._assets.values())
    

    def enabled_assets(self) -> List[Asset]:
        """
        Return all enabled loaded materials.
        """
        settings = get_aiconsole_settings()
        return [
            asset for asset in self._assets.values() if
            settings.get_asset_status(self.asset_type, asset.id) in [
                AssetStatus.ENABLED]
        ]

    def forced_assets(self) -> List[Asset]:
        """
        Return all forced loaded materials.
        """
        settings = get_aiconsole_settings()
        return [
            asset for asset in self._assets.values() if
            settings.get_asset_status(self.asset_type, asset.id) in [
                AssetStatus.FORCED]
        ]

    
    @property
    def assets_project_dir(self) -> Dict[str, Asset]:
        """
        Return all forced loaded materials.
        """
        return {
            material.id: material
            for material in self._assets.values()
            if material.defined_in == AssetLocation.PROJECT_DIR
        }
    
    @property
    def assets_aiconsole_core(self) -> Dict[str, Asset]:
        """
        Return all forced loaded materials.
        """
        return {
            material.id: material
            for material in self._assets.values()
            if material.defined_in == AssetLocation.AICONSOLE_CORE
        }
    
    async def save_asset(self, asset: Asset, new: bool, old_asset_id: Optional[str] = None):
        self._assets[asset.id] = await save_asset_to_fs(asset, new, old_asset_id)

    def delete_asset(self, asset_id):
        delete_asset_from_fs(self.asset_type, asset_id)
        del self._assets[asset_id]

    def move(self, old_asset_id: str, new_asset_id: str) -> None:
        move_asset_in_fs(self.asset_type, old_asset_id, new_asset_id)

    def get_asset(self, name):
        """
        Get a specific asset.
        """
        if name not in self._assets:
            raise KeyError(f"Asset {name} not found")
        return self._assets[name]
    
    async def reload(self, initial: bool = False):
        _log.info(f"Reloading {self.asset_type}s ...")

        self._assets = {}

        asset_ids = set([
            os.path.splitext(os.path.basename(path))[0]
            for paths_yielding_function in [
                list_files_in_file_system(get_core_assets_directory(self.asset_type)),
                list_files_in_file_system(get_project_assets_directory(self.asset_type)),
            ] for path in paths_yielding_function
            if os.path.splitext(Path(path))[-1] == ".toml"
        ])

        for id in asset_ids:
            try:
                self._assets[id] = await load_asset_from_fs(self.asset_type, id)
            except Exception as e:
                await ErrorWSMessage(
                    error=f"Invalid {self.asset_type} {id} {e}",
                ).send_to_all()
                continue

        await AssetsUpdatedWSMessage(
            initial=initial,
            asset_type=self.asset_type,
            count=len(self._assets),
        ).send_to_all()